from chaincrafter.chains import Chain
from chaincrafter.prompts import Prompt, response_format_list, response_style, response_length, extract_items_from_list

__export__ = [Chain, Prompt, response_format_list, response_style, response_length, extract_items_from_list]



from typing import Callable


class Prompt:
    """
    A chat message prompt that is formatted with input variables and modified by prompt modifiers.
    """
    def __init__(self, fmt_str: str, prompt_modifiers: [Callable] = [], **input_vars_transformers: [dict[str, Callable]]):
        """
        A prompt is a string that is formatted with input variables and modified by prompt modifiers.

        :param fmt_str: The prompt string, with format variables in the form of {var_name}
        :param prompt_modifiers: Functions that modify the prompt string
        :param input_vars_transformers: The input variables and the functions that transform them
          such as `str` or `int` or `lambda x: x.lower()`.

        Usage:
        >>> prompt = Prompt("Add {a} and {b}", a=int, b=int)
        >>> prompt = Prompt(
        ...     fmt_str="Hello {name}!",
        ...     name=lambda x: x.capitalize(),
        ...     prompt_modifiers=[lambda x: x + " How are you?"],
        ... )
        """
        self._fmt_str = fmt_str
        self._prompt_modifiers = prompt_modifiers
        self._input_vars_transformers = input_vars_transformers

    def build(self, **input_vars):
        """
        Applies the input variable transformers to the input variables, then formats the prompt string.

        :param input_vars: key-value pairs of input variables
        :return: the formatted prompt string

        Usage:
        >>> prompt = Prompt(
        ...     fmt_str="Hello {name}!",
        ...     name=lambda x: x.capitalize(),
        ...     prompt_modifiers=[lambda x: x + " How are you?"],
        ... )
        >>> prompt.build(name="john")
        'Hello John! How are you?'
        """
        for input_var, transformer in self._input_vars_transformers.items():
            input_vars[input_var] = transformer(input_vars[input_var])
        formatted = self._fmt_str.format(**input_vars)
        for prompt_modifier in self._prompt_modifiers:
            formatted = prompt_modifier(formatted)
        return formatted


# Response Format Functions
def response_format_list(example_item_text: str) -> Callable[[str], str]:
    def formatter(formatted_message):
        items = '\n'.join([f"{i + 1}: {example_item_text} {i + 1}" for i in range(3)])
        return formatted_message + f"""
Format the answer in the following way:

{items}"""
    return formatter


def response_style(style_of: str) -> str:
    def formatter(formatted_message):
        return formatted_message + f"\nRespond in the style of {style_of}"
    return formatter


def response_length(short_or_long: str = "short", answer_or_paragraph: str = "sentence") -> str:
    def formatter(formatted_message):
        return formatted_message + f"\nRespond with a {short_or_long} {answer_or_paragraph}"
    return formatter


# Response Extraction Functions
def extract_items_from_list(response: str) -> [str]:
    return [item.split(":")[1].strip() for item in response.split("\n") if item.strip()]



import asyncio
from collections import deque
import logging
from typing import TypedDict

from chaincrafter.models import ChatModel
from chaincrafter.prompts import Prompt


Message = TypedDict('Message', {'role': str, 'content': str})


class Chain:
    def __init__(self, system_prompt: Prompt, *prompts_and_output_keys: [Prompt, str]):
        """
        Sets up the system prompt and a chain of prompts and output keys.
        :param system_prompt: The prompt that sets up the system and it gives it some starting information.
        :param prompts_and_output_keys: A list of tuples of prompts and output keys. The prompts are prompts for the
            user, they specify their own input keys. The output keys are the keys that the user's input will be stored
            in.
        """
        self._system_prompt = system_prompt
        self._prompts_and_output_keys = prompts_and_output_keys

    def run(self, chat_model: ChatModel, starting_input_vars: dict = {}) -> [Message]:
        """
        Runs the chain. The output of each prompt is passed to the next prompt as a starting input variable.

        Ensure that the keys of the starting input variables and the output are unique or they will be overwritten.

        :param chat_model: The LLM chat model to use.
        :param starting_input_vars: The starting input variables that are passed to the first prompt
            (not the system prompt)
        :return: The messages that were sent back and forth between the user and the system.
        """
        queue = deque(self._prompts_and_output_keys)

        input_vars = {key: value for key, value in starting_input_vars.items()}
        logging.debug(f"Starting input variables: {input_vars}")

        messages = [
            {"role": "system", "content": self._system_prompt.build(**input_vars)}
        ]
        while queue:
            human_prompt, output_key = queue.popleft()
            messages.append({"role": "user", "content": human_prompt.build(**input_vars)})
            response = chat_model.complete(messages)

            if output_key in input_vars:
                logging.warning(f"Overwriting input variable {output_key}")

            input_vars.update({output_key: response})
            messages.append({"role": "assistant", "content": response})
        return messages

    async def async_run(self, chat_model: ChatModel, starting_input_vars: dict = {}) -> [Message]:
        return await asyncio.to_thread(self.run, chat_model, starting_input_vars)



import unittest

from chaincrafter.chains import Chain, Prompt
from chaincrafter.models import MockChat


class TestMockChat(unittest.TestCase):
    def test_prompts(self):
        model = MockChat(0.9, 'mock')
        system_prompt = "Hello world!"
        expected = [
            {
                'role': 'system',
                'content': system_prompt,
            },
            {
                'role': 'user',
                'content': 'Hello?',
            },
            {
                'role': 'assistant',
                'content': 'Yes, hello'
            },
        ]
        model.mock_responses(["Yes, hello"])
        chain = Chain(Prompt(system_prompt), (Prompt("Hello?"), "output"))
        result = chain.run(model)
        self.assertListEqual(result, expected)



from setuptools import setup, find_packages

setup(
    name="chaincrafter",
    version="0.2.0",
    description="Seamless integration and composability for large language model apps.",
    long_description="Seamless integration and composability for large language model apps.",
    author="Rudolf Olah",
    author_email="rudolf.olah.to@gmail.com",
    url="https://github.com/rudolfolah/chaincrafter",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Other/Nonlisted Topic",
    ],
)



import logging
import os
import subprocess
import sys


# Base Model
class ChatModel:
    """
    A base class for chat models.
    """
    def __init__(self, temperature: float, model_name: str):
        pass

    def complete(self, messages: [dict]) -> str:
        pass

    def show_usage(self):
        pass


# Mock Models
class MockChat(ChatModel):
    def __init__(self, temperature, model_name):
        self._temperature = temperature
        self._model_name = model_name
        self._mock_responses = []

    def mock_responses(self, messages: [str]):
        self._mock_responses = messages

    def complete(self, messages: [dict]) -> str:
        return self._mock_responses.pop(0)


# Integration Models
class LlamaCppChat(ChatModel):
    """
    A chat model that uses the llama.cpp library.

    Requires that the LLAMA_CPP_PATH environment variable is set.

    Example usage:
    >>> model = LlamaCppChat(0.9, '/path/to/model.bin')
    >>> model.complete([{ 'role': 'system', 'content': 'Hello world!' }])
    >>> model = LlamaCppChat(0.9, '/Users/rudolfo/Library/Application Support/nomic.ai/GPT4All/nous-hermes-13b.ggmlv3.q4_0.bin')
    >>> model.complete([{ 'role': 'system', 'content': 'Hello world!' }, { 'role': 'user', 'content': 'What are five different trees in North America?' }])
    """

    def __init__(self, temperature, model_name):
        """

        :param temperature:
        :param model_name: The path to the model to load
        """
        self._bin_path = os.getenv("LLAMA_CPP_PATH")
        self._temperature = temperature
        self._model_name = model_name

        # for testing
        self._bin_path = '/Users/rudolfo/Workspace/llama.cpp/main'
        self._model_name = '/Users/rudolfo/Library/Application Support/nomic.ai/GPT4All/nous-hermes-13b.ggmlv3.q4_0.bin'

    def complete(self, messages):
        _process: subprocess.Popen = subprocess.Popen(
            [self._bin_path, '--model', self._model_name, '--interactive-first', '--temp', str(self._temperature), '-ngl', '1'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        starting_prompt = rf"""Transcript of a dialog between a user and an assistant named System. System is a person who is helpful, kind, honest and provides immediate and precise answers without having to do any additional research. {messages[0]['content']}\
User: Hello, System.\
System: Hello! How can I assist you today?\
User: {messages[1]['content']}\
System: /"""
        _process.communicate(input=(starting_prompt).encode())
        print(_process.communicate())


class OpenAiChat(ChatModel):
    """
    A chat model that uses OpenAI's API.

    Requires that the OPENAI_API_KEY environment variable is set.
    """
    def __init__(
            self,
            temperature: float,
            model_name: str,
            top_p: float = 1.0,
            n: int = 1,
            stream: bool = False,
            stop: [str] = None,
            max_tokens: int = None,
            presence_penalty: float = 0.0,
            frequency_penalty: float = 0.0,
            logit_bias: dict = None,
    ):
        super().__init__(temperature, model_name)
        import openai
        openai.api_key = os.getenv("OPENAI_API_KEY")
        if openai.api_key is None:
            logging.error("OPENAI_API_KEY environment variable is not set")
            sys.exit(1)
        self._temperature = temperature
        self._model_name = model_name
        self._top_p = top_p
        self._n = n
        self._stream = stream
        self._stop = stop
        self._max_tokens = max_tokens
        self._presence_penalty = presence_penalty
        self._frequency_penalty = frequency_penalty
        self._logit_bias = logit_bias
        logging.info(f"Using OpenAI model {self._model_name}, temperature {self._temperature}")
        self.usage = {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
        }

    def complete(self, messages: [dict]):
        """
        Completes a chat using OpenAI's API.

        :param messages: A list of messages to send to the chat model
        :return:
        """
        import openai
        logging.debug("Calling OpenAI API")
        kwargs = {
            "model": self._model_name,
            "temperature": self._temperature,
            "messages": messages,
            "top_p": self._top_p,
            "n": self._n,
            "stream": self._stream,
            "presence_penalty": self._presence_penalty,
            "frequency_penalty": self._frequency_penalty,
        }
        if self._stop:
            kwargs["stop"] = self._stop
        if self._max_tokens:
            kwargs["max_tokens"] = self._max_tokens
        if self._logit_bias:
            kwargs["logit_bias"] = self._logit_bias
        completion = openai.ChatCompletion.create(**kwargs)
        logging.debug("OpenAI API call complete")
        self.usage["prompt_tokens"] += completion.usage["prompt_tokens"]
        self.usage["completion_tokens"] += completion.usage["completion_tokens"]
        self.usage["total_tokens"] += completion.usage["total_tokens"]
        return completion.choices[0].message["content"]

    def show_usage(self):
        print(f"Prompt tokens: {self.usage['prompt_tokens']}")
        print(f"Completion tokens: {self.usage['completion_tokens']}")
        print(f"Total tokens: {self.usage['total_tokens']}")



# Based on the API from prompttools: https://prompttools.readthedocs.io/en/latest/experiment.html
import copy
from datetime import datetime
from itertools import product

import pandas as pd

from chaincrafter.chains import Chain
from chaincrafter.models import OpenAiChat


class Experiment:
    """
    An experiment that runs a chat model on a prompt. Generates all permutations of the variables provided.
    """

    model_class = None

    def __init__(self, chain: Chain, **model_params):
        self.chain = chain
        self.model_params = model_params
        self._model_params_product = None
        self._prepared = False
        self.results = []
        self.starting_input_vars = {}

    def set_chain_starting_input_vars(self, starting_input_vars: dict) -> None:
        """
        Sets the starting input variables for the chain.
        :param starting_input_vars:
        """
        self.starting_input_vars = copy.deepcopy(starting_input_vars)

    def _set_model_params(self, default_params: dict) -> None:
        """
        Sets default values for model parameters that are not specified in the experiment.
        :param default_params: A dict of default parameter values
        """
        for key, value in default_params.items():
            if key not in self.model_params:
                self.model_params[key] = value

    def prepare(self) -> None:
        """
        Prepares the experiment by generating all permutations of the model parameters.
        """
        self._model_params_product = product(*self.model_params.values())
        self._prepared = True

    def run(self, runs: int = 1) -> None:
        """
        Runs the experiment. Stores the results in self.results including the timestamp, run number, and messages and the model parameters.
        :param runs: The number of times to run the experiment
        """
        if not self._prepared:
            self.prepare()
        for run_number in range(runs):
            for model_params_values in self._model_params_product:
                model_params = dict(zip(self.model_params.keys(), model_params_values))
                model = self.model_class(**model_params)
                messages = self.chain.run(model, self.starting_input_vars)
                result = {
                    "run_number": run_number,
                    "timestamp": datetime.now().isoformat(),
                    "messages": messages,
                }
                result.update(model_params)
                self.results.append(result)
                print(f"Run {run_number} at {result['timestamp']} with model parameters: {model_params}")
                for message in messages:
                    print(f"{message['role']}: {message['content']}")
                print()

    def to_csv(self):
        return pd.DataFrame(self.results).to_csv()

    def to_pandas_df(self):
        return pd.DataFrame(self.results)

    def to_json(self):
        return pd.DataFrame(self.results).to_json()

    def visualize(self):
        return pd.DataFrame(self.results)


class OpenAiChatExperiment(Experiment):
    model_class = OpenAiChat

    def __init__(self, chain, **model_params):
        super().__init__(chain, **model_params)
        self._set_model_params({
            "temperature": [1.0],
            "top_p": [1.0],
            "n": [1],
            "stream": [False],
            "stop": [None],
            "max_tokens": [None],
            "presence_penalty": [0.0],
            "frequency_penalty": [0.0],
            "logit_bias": [None],
        })
